/*

Copyright (C) 2003 J.D. Cole <jd@transientresearch.com>

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

$Id: oct-pi.cc,v 1.8.2.1 2003/03/03 22:23:58 jd Exp $
*/

#if defined (__GNUG__) && defined (USE_PRAGMA_INTERFACE_IMPLEMENTATION)
#pragma implementation
#endif

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "error.h"
#include "oct-pi.h"
#include "utils.h"

// !! It would be better to use the Octave shared library facilities
#include <dlfcn.h>

#include <assert.h>

// If TRUE, user wants to start parallel libraries
// (--parallel) for compatability
bool parallel_flag;

octave_parallel_interface parallel_interface;

octave_parallel_interface::octave_parallel_interface (void)
  : pil_handle(0), current_error_state(no_error), 
  current_error_string("No error"), state(pre_initialize), 
  initialize_fcn(0), finalize_fcn(0), error_string_fcn(0)
{
}

octave_parallel_interface::~octave_parallel_interface (void)
{
  // Finalization should do all deallocation necessary on the
  // "other" side

  // Close the dynamic library
}

pi_error
octave_parallel_interface::initialize (int *argc, char **argv[])
{
  current_error_state = no_error;
  state = pre_initialize;

  // 1. Is --parallel on the command line?
  bool parallel_run = false;

  std::string arg_str;

  for (int i = 1; i< *argc; i++)
    {
      arg_str = (*argv)[i];

      if (arg_str.substr(0,10) == "--parallel")
	parallel_run = true;
    }

  if (! parallel_run)
    return no_error;

  // 2. Check for a --parallel [file] if its not there, look for "paralleloctave.pil",
  // the PIL library, and save its name.

  // !! Octave isn't liking optional arguments, so right now it required
  std::string default_pil_name= "paralleloctave";

  if (! arg_str.substr(10).empty ())
    {
      // String the leading white space and pull out the 
      std::string::size_type path_start = arg_str.substr(10).find_first_not_of(" ");

      default_pil_name = arg_str.substr(path_start);

      if (default_pil_name.empty ())
	{
	  current_error_state = failed;
	  
	  current_error_string = "invalid PIL name or path";

	  return current_error_state;
	}
    }

  pil_full_path_name = file_in_path (default_pil_name,".pil");

  if (pil_full_path_name.empty ())
    {
      pil_full_path_name = file_in_path (default_pil_name,"");
    }

  if (pil_full_path_name.empty ())
    {
      current_error_state = failed;
      current_error_string = "couldn't find PIL file: " + default_pil_name;

      return current_error_state;
    }

  // 3. Verification:
  //   a. Verify PIL can be opened.
  if (open (pil_full_path_name) == failed)
    {
      current_error_state = failed;
      state = failed_to_load;

      current_error_string = "failed to open PIL file:" + default_pil_name;

      return current_error_state;
    }

  //   b. Verify and save refereces to the PIL functions:
  //      (!! does C++ mangle C names it compiles ?)
  //      initialize 
  //      post_parse 
  //      eval_section_list
  //      finalize   
  //      error_string
  
  // !! Should *_fcn_type be renamed to pil*_fcn_type?
  initialize_fcn =	  
    (pil_initialize_fcn_type)        search ("pil_initialize");

  install_builtins_fcn =  
    (pil_install_builtins_fcn_type)  search ("pil_install_builtins");

  finalize_fcn =	
    (pil_finalize_fcn_type)          search ("pil_finalize");

  error_string_fcn =
    (pil_error_string_fcn_type)      search ("pil_error_string");

  // 5. Call initialization function, returning any error state that exists
  current_error_state = initialize_fcn (argc, argv);

  if (current_error_state == no_error)
    state = started;
  else
    {
      // !! Set the running state
      state = failed_to_initialize;
      
      current_error_state = failed;
      
      close ();
    }

  // 6. return error state

  return current_error_state;
}

pi_error
octave_parallel_interface::install_builtins (void)
{
  if (state == started)
    {
      current_error_state = install_builtins_fcn ();
      
      if (current_error_state != no_error)
	set_pil_error_string("PIL builtins install failed");
    }
  else
    {
      current_error_state = failed;

      current_error_string = "PIL can't install builtins if state is non-start";
    }

  return current_error_state;
}

pi_error
octave_parallel_interface::finalize (void)
{
  // !! Why does octave call do_octave_atexit twice??
  if (state == started)
    {
      current_error_state = finalize_fcn ();

      if (current_error_state != no_error)
	set_pil_error_string("PIL failed to finalize");
      else
	state = stopped;
    }
  else
    {
      current_error_state = failed;

      current_error_string = "PIL tried to finalize a non stopped process";
    }

  return current_error_state;
}

void
octave_parallel_interface::set_pil_error_string (const std::string msg)
{
  // !! This could be moved into .h and made one liner inlines
  if (error_string_fcn)
    current_error_string = msg + ": " + error_string_fcn ();
  else
    current_error_string = msg;
}

pi_error
octave_parallel_interface::open (const std::string pil_name)
{
  //It is a programming error to come in here twice
  assert (! pil_handle);

  pil_handle = dlopen (pil_name.c_str (), RTLD_LAZY);

  if (pil_handle)
    return no_error;
  else
    return failed;
}

void*
octave_parallel_interface::search (const std::string fcn_name)
{
  //It is a programming error to come in here twice
  assert (pil_handle);

  return (dlsym(pil_handle, fcn_name.c_str ()));
}

void
octave_parallel_interface::close (void)
{
  //It is a programming error to come in here twice
  assert (pil_handle);
  
  dlclose (pil_handle);
}
